/* device.cpp
   
   Implementation of the Device class

   Copyright (c) 2003-2008 Tim Stadelmann

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */


/** HEADER FILES **/

/* standard C++ header files */
#include <cstdio>
#include <cerrno>
using namespace std;

/* KDE header files */
#include <klocale.h>
#include <kmessagebox.h>

/* Qt header files */
#include <qstring.h>

/* program specific header files */
#include "device.hpp"


/** FUNCTIONS **/

/* The constructor for the Device class.  Apart from the parent widget
   (for dialog boxes) his function takes the number of the IDE
   controller, the path of the udevadm utility, and the path to the
   hotswap backend as arguments.  */
Device::Device (QWidget *parent_arg,
		const unsigned int ide_controller_arg,
		QString udevadm_path_arg,
		QString hotswap_path_arg)
{
  parent = parent_arg;
  ide_controller = ide_controller_arg;
  udevadm_path = udevadm_path_arg;
  hotswap_path = hotswap_path_arg;

  /* Probe the currently configured device and initialize the internal
     state.  */
  probe ();
}


/* This method returns true if a device is present.  */
bool
Device::present ()
{
  /* Return the internal flag indicating the presence of a device.  */
  return present_flag;
}

/* This method returns the IDE model string or the configured device,
   or i18n ("absent") if no device is configured.  */
QString
Device::name ()
{
  return model_string;
}


/* Call the backend to find out whether there are any mounted
   filesystems left on the device.  Return true if there are mounted
   filesystems or an error occurred, false otherwise.  */
bool
Device::mounted ()
{
  bool mounted;			// whether filesystems are mounted

  /* Run the backend.  */
  if (!call_backend (" -n mounted-ide"))
    {
      mounted = true;
    }
  else
    {
      /* Check the result and return the appropriate value.  */
      if (strncmp (i18n ("no"), buffer, strlen (i18n ("no"))) == 0)
	  mounted = false;
      else
	  mounted = true;
    }

  return mounted;
}


/* Probe the installed device.  */
void
Device::probe ()
{
  /* Call the backend with the appropriate options. */
  if (call_backend (" -n probe-ide"))
    {
      /* Only the side effect of the strtok function is used.  */
      strtok (buffer, "\n");

      /* Save the result in the model_string variable.  */
      strncpy (model_string, buffer, 40);
      
      /* Check whether the result indicates that no device is present,
	 and set the flag accordingly.  */
      if (strcoll ("absent", model_string) == 0)
	  present_flag = false;
      else
	  present_flag = true;
    }
  else
    {
      strncpy (model_string, "[error]", 40);
      present_flag = false;
    }
}


/* Call the backend with the appropriate arguments for rescanning the
   IDE bus.  */
void
Device::rescan ()
{
  for (;;)
    {
      /* Check whether a device is already present.  This can happen
         if another program instructed the kernel to rescan the IDE
         bus behind our back, or if there is a delay between
         rescanning the bus and new devices becoming visible.  */
      probe ();
      if (present_flag)
	{
	  KMessageBox::error
	    (parent,
	     i18n ("A device is already present."),
	     i18n ("Backend Error"));
	  break;
	}
      
      /* Call the backend to rescan the IDE bus.  */
      call_backend (" -n rescan-ide");

      /* If no new device could be found, ask the user whether to
         abort or try again.  */
      probe ();
      if (!present_flag)
	{      
	  if (KMessageBox::warningYesNo
	      (parent,
	       i18n ("No IDE device has been found.  If you are sure that "
		     "the device connects to the IDE bus, verify that the "
		     "module is inserted correctly.\n"
		     "Do you want to try again?"),
	       i18n ("No Device Found"))
	      == KMessageBox::No)
	    {
	      probe ();
	      break;
	    }
	}
      else
	{
	  break;
	}
    }
}


/* Call the backend with the appropriate arguments for unregistering
   the device.  */
void
Device::unregister ()
{
  /* Check whether the device is really present.  */
  probe ();
  if (!present_flag)
    {
      KMessageBox::error
	(parent,
	 i18n ("There is no device to remove."),
	 i18n ("Backend Error"));
    }
  else
    {
      /* Check whether filesystems on the device are still
	 mounted.  */
      if (mounted ())
	{
	  KMessageBox::error
	    (parent,
	     i18n ("At least one filesystem on the "
		   "current device is still mounted.\n"
		   "Unmount the mounted filesystems "
		   "and try again."),
	     i18n ("Mounted Filesystem"));
	}
      else
	{
	  /* Call the backend to remove the device.  */
	  if (!call_backend (" -n unregister-ide"))
	    {
	      KMessageBox::error
		(parent,
		 i18n ("Unregistering the IDE device failed."),
		 i18n ("Backend Error"));
	    }
	  probe ();
	}
    }
}


/* Call the backend with arbitrary arguments and deal with errors.
   Read both stdout and stderr of the backend into buffer.  Return
   false if an error has occured, true otherwise.  */

bool
Device::call_backend (QString arguments)
{
  /* Construct the command string passed to the shell.  */

  QString command;		// the command string
  command += hotswap_path;
  command += " --ide-controller " + QString::number (ide_controller);
  command += " --udevadm-path " + udevadm_path;
  command += " " + arguments;
  command += " 2>&1";

  /* Run the backend.  */

  FILE *backend_out;		// the output stream from the backend
  backend_out = popen (command.latin1 (), "r");
  if (backend_out == NULL)
    {      
      QString msg;		// the error message
      msg = i18n ("An error occurred while calling the backend:\n");
      msg += strerror (errno);
      KMessageBox::error (parent, msg, i18n ("Backend Error"));

      return false;
    }

  /* Read the output of the backend.  Surprisingly enough, this cannot
     be done in a standard way using the C++ stream classes.  */
  while(fgets (buffer, BUFFER_LENGTH, backend_out));

  /* Close the connection to the backend.  pclose does return the exit
     code of the program if it has succeeded.  The fact that at least
     one version of the Linux man page fails to mention this is a bug
     in the documentation, not in the library.  */
  int ret_value;		// the return value of pclose
  ret_value = pclose (backend_out);

  if (ret_value == -1)
    {
      /* In this case, the pclose call itself failed.  */
      QString msg;		// the error message
      msg = i18n ("An error occurred while calling the backend:\n");
      msg += strerror (errno);
      KMessageBox::error (parent, msg, i18n ("Backend Error"));
    }
  else if (ret_value)
    {
      /* Any other return value corresponds to the exit status of the
	 backend.  */
      KMessageBox::error
	(parent,
	 i18n ("An error occurred while running the backend."),
	 i18n ("Backend Error"));
    }

  return ret_value != -1;
}
