/* message_box.c

   Utility functions for handling Motif message dialog boxes.

   Copyright (C) 2001-2003 Tim Stadelmann

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */


/** HEADER FILES**/

/* autoconf */
#if HAVE_CONFIG_H
#  include <config.h>
#endif

/* Motif header files */
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/MwmUtil.h>

/* ISO C header files */
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

#if ENABLE_NLS
#  include <libintl.h>
#  define _(Text) gettext (Text)
#else
#  define _(Text) Text
#endif

/* program specific header files */
#include "message_box.h"


/** MACROS **/

/* the block size used when incrementally allocating storage for
   strings */
#define BLOCKSIZE (256)
/* the size of the resources array */
#define ARGS_LENGTH (10)


/** FUNCTION PROTOTYPES **/

static answer_t vmessage (Widget parent, dialog_type_t type,
			  const char *format,
			  va_list argument_list);
static Widget create_dialog (Widget parent, dialog_type_t type);
static void callback (Widget, XtPointer, XtPointer);
static int msgbox_vprintf (Widget msg_box, const char *format,
			   va_list argument_list);
static XmString wrap_string (XmRenderTable render_table,
			     unsigned int width,
			     const char *string);

/** FUNCTIONS **/

/* Display a formatted string in a Motif question dialog with "Yes"
   and "No" buttons, wait for the user to interact with it, and return
   a value indicating the user's response.  The default button is
   "Yes"; pressing ESC or closing the window is equivalent to pressing
   the "No" button.  See the comment for message for more details.  */
answer_t
question_yes_no (Widget parent, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, QUESTION_YES_NO, format,
		     argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif question dialog with "Yes",
   "No", and "Cancel" buttons, wait for the user to interact with it,
   and return a value indicating the user's response.  The default
   button is "Yes"; pressing ESC or closing the window is equivalent
   to pressing the "Cancel" button.  See the comment for message for
   more details.  */
answer_t
question_yes_no_cancel (Widget parent, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, QUESTION_YES_NO_CANCEL, format,
		     argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif warning dialog with "Yes" and
   "No" buttons, wait for the user to interact with it, and return a
   value indicating the user's response.  The default button is "No";
   pressing ESC or closing the window is equivalent to pressing the
   "No" button.  See the comment for message for more details.  */
answer_t
warning_yes_no (Widget parent, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, WARNING_YES_NO, format,
		     argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif warning dialog with
   "Continue" and "Cancel" buttons, wait for the user to interact with
   it, and return a value indicating the user's response.  The default
   button is "Continue"; pressing ESC or closing the window is
   equivalent to pressing the "Cancel" button.  See the comment for
   message for more details.  */
answer_t
warning_continue_cancel (Widget parent, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, WARNING_CONTINUE_CANCEL, format,
		     argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif warning dialog with "Yes",
   "No", and "Cancel" buttons, wait for the user to interact with it,
   and return a value indicating the user's response.  The default
   button is "Yes"; pressing ESC or closing the window is equivalent
   to pressing the "No" button.  See the comment for message for more
   details.  */
answer_t
warning_yes_no_cancel (Widget parent, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, WARNING_YES_NO_CANCEL, format,
		     argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif information dialog and wait
   for the user to acknowledge it.  See the comment for message for
   more details.  */
answer_t
information (Widget parent, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, INFORMATION, format,
		     argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif error dialog and wait for the
   user to acknowledge it.  See the comment for message for more
   details.  */
answer_t
error (Widget parent, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, ERROR, format,
		     argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif message dialog box of the
   indicated type, wait for the user to interact with it, and return a
   value indicating the user's response (see message_box.h for
   possible values).  The dialog shell is a popup child of the parent
   widget.  The format and variable arguments are interpreted as by
   printf and friends, and the resulting string is wrapped to a
   quarter of the screen width.  The function implements its own
   message loop and will not return until a response from the user has
   been obtained.  */
answer_t
message (Widget parent, dialog_type_t type, const char *format, ...)
{
  va_list argument_list;	/* the argument list */
  answer_t answer;		/* the answer from the user */
  
  va_start (argument_list, format);
  answer = vmessage (parent, type, format, argument_list);
  va_end (argument_list);
  return answer;
}

/* Display a formatted string in a Motif message dialog box of the
   indicated type, wait for the user to interact with it, and return a
   value indicating the user's response.  The dialog shell is a popup
   child of the parent widget.  The format and variable arguments are
   interpreted as by vprintf and friends, and the resulting string is
   wrapped to a quarter of the screen width.  The function implements
   its own message loop and will not return until a response from the
   user has been obtained.  */
static answer_t
vmessage (Widget parent, dialog_type_t type, const char *format,
	  va_list argument_list)
{
  answer_t answer;		/* the response from the user */
  XtAppContext app_context;	/* the application context */
  Widget msg_box;		/* the dialog box */
  Arg args[ARGS_LENGTH];	/* an array for defining resources */
  Cardinal n;			/* the number of  */
  char *strtbl[] =  {"UNDEFINED", "OK", "CANCEL",
	     "YES", "NO", "CONTINUE"};
  
  /* Get the handle to the dialog box.  */
  msg_box = create_dialog (parent, type);
  
  /* Set the message string.  */
  msgbox_vprintf (msg_box, format, argument_list);

  /* Manage the dialog.  */
  XtManageChild (msg_box);

  /* Set the user data of the dialog box to the address of the local
     variable answer.  This variable can then be accessed and changed
     by the callback.  */
  n = 0;
  XtSetArg (args[n], XmNuserData, &answer), n++;
  assert (n < ARGS_LENGTH);
  XtSetValues (msg_box, args, n);

  answer = UNDEFINED;
  app_context = XtWidgetToApplicationContext (msg_box);

  /* Keep processing events until an answer has been obtained from the
     user.  */
  while (answer == UNDEFINED)
    {
      XEvent event;
      XtAppNextEvent (app_context, &event);
      XtDispatchEvent (&event);
    }

  /* Destroy the dialog box.  */
  XtDestroyWidget (msg_box);
  
  return answer;
}

/* Create a message box widget of the required type, set the
   appropriate callbacks, and return a handle to it.  */
static Widget
create_dialog (Widget parent, dialog_type_t type)
{
  /* A pointer to the appropriate widget creation function.  */
  Widget (*create_widget) (Widget, char *, Arg *, Cardinal);

  Widget msg_box;		/* the dialog box */
  Widget button;		/* the additional button */
  XmString title;		/* the title of the dialog box */
  XmString ok_label;		/* the text of the OK button */
  answer_t ok_answer;		/* the answer for the OK button */
  Boolean use_cancel;		/* whether to use the cancel button */
  XmString cancel_label;	/* the text of the cancel button */
  answer_t cancel_answer;	/* the answer for the cancel button */
  Boolean use_button;		/* whether to use the extra button */
  XmString button_label;	/* the text of the additional button */
  answer_t button_answer;	/* the answer for the extra button */
  unsigned int default_button;	/* the default button type*/
  Arg args[ARGS_LENGTH];	/* an array for defining resources */
  Cardinal n;			/* the number of valid resources */
  
  /* Initalize the variables used in creating the message box
     according to the required type and button configuration.  */
  switch (type)
    {
    case QUESTION_YES_NO:
      create_widget = &XmCreateQuestionDialog;
      title = XmStringCreateLocalized (_("Question"));
      ok_label = XmStringCreateLocalized (_("Yes"));
      ok_answer = YES;
      use_cancel = True;
      cancel_label = XmStringCreateLocalized (_("No"));
      cancel_answer = NO;
      use_button = False;
      default_button = XmDIALOG_OK_BUTTON;
      break;
    case QUESTION_YES_NO_CANCEL:
      create_widget = &XmCreateQuestionDialog;
      title = XmStringCreateLocalized (_("Question"));
      ok_label = XmStringCreateLocalized (_("Yes"));
      ok_answer = YES;
      use_cancel = True;
      cancel_label = XmStringCreateLocalized (_("Cancel"));
      cancel_answer = CANCEL;
      use_button = True;
      button_label = XmStringCreateLocalized (_("No"));      
      button_answer = NO;
      default_button = XmDIALOG_OK_BUTTON;
      break;
    case WARNING_YES_NO:
      create_widget = &XmCreateWarningDialog;
      title = XmStringCreateLocalized (_("Warning"));
      ok_label = XmStringCreateLocalized (_("Yes"));
      ok_answer = YES;
      use_cancel = True;
      cancel_label = XmStringCreateLocalized (_("No"));
      cancel_answer = NO;
      use_button = False;
      default_button = XmDIALOG_CANCEL_BUTTON;
      break;
    case WARNING_CONTINUE_CANCEL: 
      create_widget = &XmCreateWarningDialog;
      title = XmStringCreateLocalized (_("Warning"));
      ok_label = XmStringCreateLocalized (_("Continue"));
      ok_answer = CONTINUE;
      use_cancel = True;
      cancel_label = XmStringCreateLocalized (_("Cancel"));
      cancel_answer = CANCEL;
      use_button = False;
      default_button = XmDIALOG_OK_BUTTON;
      break;
    case WARNING_YES_NO_CANCEL:
      create_widget = &XmCreateWarningDialog;
      title = XmStringCreateLocalized (_("Question"));
      ok_label = XmStringCreateLocalized (_("Yes"));
      ok_answer = YES;
      use_cancel = True;
      cancel_label = XmStringCreateLocalized (_("Cancel"));
      cancel_answer = CANCEL;
      use_button = True;
      button_label = XmStringCreateLocalized (_("No"));      
      button_answer = NO;
      default_button = XmDIALOG_OK_BUTTON;
      break;
    case INFORMATION:
      create_widget = &XmCreateInformationDialog;
      title = XmStringCreateLocalized (_("Information"));
      ok_label = XmStringCreateLocalized (_("OK"));
      ok_answer = OK;
      use_cancel = False;
      cancel_answer = OK;	/* needed if the user presses ESC */
      use_button = False;
      default_button = XmDIALOG_OK_BUTTON;
      break;
    case ERROR:
      create_widget = &XmCreateErrorDialog;
      title = XmStringCreateLocalized (_("Error"));
      ok_label = XmStringCreateLocalized (_("OK"));
      ok_answer = OK;
      use_cancel = False;
      cancel_answer = OK;	/* needed if the user presses ESC */
      use_button = False;
      default_button = XmDIALOG_OK_BUTTON;
      break;
    }

  /* Create the message box.  */
  n = 0;
  XtSetArg (args[n], XmNokLabelString, ok_label), n++;
  if (use_cancel)
    XtSetArg (args[n], XmNcancelLabelString, cancel_label), n++;
  XtSetArg (args[n], XmNdialogTitle, title), n++;
  XtSetArg (args[n], XmNdefaultButtonType, default_button), n++;
  XtSetArg (args[n], XmNdialogStyle,
	    XmDIALOG_PRIMARY_APPLICATION_MODAL), n++;
  XtSetArg (args[n], XmNmwmDecorations,
	    MWM_DECOR_ALL | MWM_DECOR_RESIZEH), n++;
  assert (n < ARGS_LENGTH);
  msg_box = (*create_widget) (parent, "message_box", args, n);
  XmStringFree (ok_label);
  if (use_cancel)
    XmStringFree (cancel_label);
  XmStringFree (title);

  /* Add message box callbacks.  */
  XtAddCallback (msg_box, XmNokCallback, callback,
		 (XtPointer) ok_answer);    
  XtAddCallback (msg_box, XmNcancelCallback, callback,
		 (XtPointer) cancel_answer);    
  XtAddCallback (msg_box, XmNunmapCallback, callback,
		 (XtPointer) cancel_answer);    

  /* Unmanage unneeded buttons and create an extra button if
     required.  */
  XtUnmanageChild (XtNameToWidget (msg_box, "Help"));
  if (!use_cancel)
    XtUnmanageChild (XtNameToWidget (msg_box, "Cancel"));
  if (use_button)
    {
      n = 0;
      XtSetArg (args[n], XmNlabelString, button_label), n++;
      button = XmCreatePushButtonGadget (msg_box, "button", args, n);
      XmStringFree (button_label);
      XtManageChild (button);
      XtAddCallback (button, XmNactivateCallback, callback,
		     (XtPointer) button_answer);
    }
  
  return msg_box;
}

/* The callback routine.  Retrieve the address of the answer variable
   in the relevant instance of vmessage and set it according to
   the button pressed by the user.  */
static void
callback (Widget widget, XtPointer client_data,
	  XtPointer call_data)
{
  answer_t *answer_ptr;		/* the address of the answer */
  Arg args[ARGS_LENGTH];	/* an array for defining resources */
  Cardinal n;			/* the number of valid resources */

  XmAnyCallbackStruct *call = (XmAnyCallbackStruct *) call_data;

  /* If widget refers to one of the buttons the message box is
     obtained by finding its parent.  */
  if (XmIsPushButtonGadget (widget))
    widget = XtParent (widget);
  assert (XmIsMessageBox (widget));

  /* Find a pointer to the answer variable and set it according to the
     information provided in client_data.  */
  n = 0;
  XtSetArg (args[n], XmNuserData, &answer_ptr), n++;
  assert (n < ARGS_LENGTH);
  XtGetValues (widget, args, n);
  *answer_ptr = (answer_t) client_data;
}

/* Format a string according to the conventions used by printf.  Wrap
   it to a quarter of the screen width using the render table of the
   label of the msg_box.  Finally configure this label to display the
   resulting string.  This function simply calls msgbox_vprinf below
   with the appropriate arguments.  */
int
msgbox_printf (Widget msg_box, const char *format, ...)
{
  int n_chars;			/* the number of characters printed */
  va_list argument_list;	/* the argument list */

  /* Set the message string.  */
  va_start (argument_list, format);
  n_chars = msgbox_vprintf (msg_box, format, argument_list);
  va_end (argument_list);

  return n_chars;
}

/* Format a string according to the conventions used by vprintf.  Wrap
   it to a quarter of the screen width using the render table of the
   label of the msg_box.  Finally configure this label to display the
   resulting string.  */
int
msgbox_vprintf (Widget msg_box, const char *format,
		va_list argument_list)
{
  XmString label;		/* for creating the compound string */
  int n_chars;			/* the return value of vsnprintf */
  char *message;		/* the message */
  int message_length;		/* the actual length of the buffer */
  XmRenderTable render_table;   /* the render table of the label */
  Arg args[ARGS_LENGTH];	/* an array for defining resources */
  Cardinal n;			/* the number of valid resources */

  message = XtMalloc (BLOCKSIZE);
  message_length = BLOCKSIZE;

  /* There are two different conventions for the return value of
     snprintf and related functions.  Please refer to the
     documentation.  */
  for (;;)
    {
      n_chars = vsnprintf (message, message_length, format,
			   argument_list);
      if (n_chars == -1)
	{
	  message_length += BLOCKSIZE;
	}
      else if (n_chars != strlen (message))
	{
	  message_length = n_chars + 1;
	}
      else
	{
	  break;
	}
      message = XtRealloc (message, message_length);
    }

  /* Retrieve the render table of the label gadget created by the
     message box.  The font information contained in it is required
     for wrapping the message to a specified width.  */
  n = 0;
  XtSetArg (args[n], XmNrenderTable, &render_table), n++;
  assert (n < ARGS_LENGTH);
  XtGetValues (XtNameToWidget (msg_box, "Message"), args, n);

  /* Create the compound string from the message, wrapping it to a
     quarter of the screen width.  */
  label = wrap_string (render_table,
		       WidthOfScreen(XtScreen (msg_box)) / 4,
		       message);
  XtFree (message);

  n = 0;
  XtSetArg (args[n], XmNmessageString, label), n++;
  assert (n < ARGS_LENGTH);
  XtSetValues (msg_box, args, n);

  XmStringFree (label);

  return n_chars;
}

static XmString
wrap_string (XmRenderTable render_table, unsigned int width,
	     const char *string)
{
  char *buffer;			/* the working buffer */
  size_t buffer_length;		/* its length */
  XmString compound_string;	/* for creating compound strings */
  unsigned int i;		/* an index into the source */
  unsigned int j;		/* an index into the buffer */
  unsigned int j_line;		/* the start of the current line */
  unsigned int j_break;		/* the most recent possible break */
  Arg args[ARGS_LENGTH];	/* an array for defining resources */
  Cardinal n;			/* the number of valid resources */

  buffer = XtMalloc (BLOCKSIZE);
  buffer_length = BLOCKSIZE;

  i = j = 0;
  j_line = 0;
  j_break = 0;

  while (string[i])
    {      
      /* Copy the next word, if any, and terminate the buffer with a
	 '\0' character.  */
      while (string[i] && !isspace (string[i]))
	{
	  /* Substract one because indices are zero-based.  Substract
	     a further two to reserve space for the "\0", "\n", or
	     "\n\n" that could be appended.  */
	  if (j == buffer_length - 1 - 2)
	    {
	      buffer_length += BLOCKSIZE;
	      buffer = XtRealloc (buffer, buffer_length);
	    }
	  buffer[j++] = string[i++];
	}
      buffer[j] = '\0';

      /* Check whether the current line exceeds the width.  If it
	 does, see whether it contains at least one possible line
	 break.  If this is the case, go back to the last one and
	 start a new line.  Otherwise accept the long line and start a
	 new line immediately.  */
      compound_string = XmStringCreateLocalized (buffer + j_line);
      if (XmStringWidth (render_table, compound_string) > width)
	{
	  if (j_break > j_line)
	    {
	      i -= j - j_break;
	      j = j_break;
	      while (isspace (string[i]))
		i++;
	      buffer[j++] = '\n';
	      j_line = j;
	    }
	  else
	    {
	      while (isspace(string[i]))
		i++;
	      buffer[j++] = '\n';
	      j_line = j;
	    }
	}

      /* If there is at least one newline character in the original
	 string, skip to the next non-empty line and insert two
	 newlines in the working buffer.  */
      if (string[i] == '\n')
	{
	  while (string[i] == '\n')
	    i++;
	  buffer[j++] = '\n';
	  buffer[j++] = '\n';
	  buffer[j] = '\0';
	  j_line = j;
	}
      
      /* Store the current position as a possible break and copy the
	 remaining white space.  */
      j_break = j;
      while (isspace (string[i]))
	{
	  /* Substract one because indices are zero-based.  */
	  if (j == buffer_length - 1)
	    {
	      buffer_length += BLOCKSIZE;
	      buffer = XtRealloc (buffer, buffer_length);
	    }
	  buffer[j++] = string[i++];
	}
    }

  /* Remove line breaks at the end of the buffer.  */
  while (buffer[--j] == '\n') buffer[j] = '\0';

  /* Create and return the final compound string.  */
  compound_string = XmStringCreateLocalized (buffer);
  XtFree (buffer);
  return compound_string;
}
