/* message_box.h

   Utility functions for handling Motif message dialog boxes.

   Copyright (C) 2003 Tim Stadelmann

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */

#ifndef DISPLAY_ERROR_H
#define DISPLAY_ERROR_H


/** STRUCTURES AND TYPES **/

/* For representing the type of message box required.  Since we also
   want to encode button configurations, the standard Motif constants
   are not sufficient.  */
typedef enum {
  QUESTION_YES_NO,
  QUESTION_YES_NO_CANCEL,
  WARNING_YES_NO,
  WARNING_CONTINUE_CANCEL,
  WARNING_YES_NO_CANCEL,
  INFORMATION,
  ERROR
} dialog_type_t;


/* For representing the response from the user.  UNDEFINED signifies
   that no response has been obtained yet.  */
typedef enum {
  UNDEFINED,
  OK,
  CANCEL,
  YES,
  NO,
  CONTINUE
} answer_t;


/** FUNCTION DECLARATIONS **/

/* Display a formatted string in a Motif question dialog with "Yes"
   and "No" buttons, wait for the user to interact with it, and return
   a value indicating the user's response.  The default button is
   "Yes"; pressing ESC or closing the window is equivalent to pressing
   the "No" button.  See the comment for message for more details.  */

extern answer_t question_yes_no (Widget parent,
				 const char *format, ...);


/* Display a formatted string in a Motif question dialog with "Yes",
   "No", and "Cancel" buttons, wait for the user to interact with it,
   and return a value indicating the user's response.  The default
   button is "Yes"; pressing ESC or closing the window is equivalent
   to pressing the "Cancel" button.  See the comment for message for
   more details.  */

extern answer_t question_yes_no_cancel (Widget parent,
					const char *format, ...);


/* Display a formatted string in a Motif warning dialog with "Yes" and
   "No" buttons, wait for the user to interact with it, and return a
   value indicating the user's response.  The default button is "No";
   pressing ESC or closing the window is equivalent to pressing the
   "No" button.  See the comment for message for more details.  */

extern answer_t warning_yes_no (Widget parent,
				const char *format, ...);


/* Display a formatted string in a Motif warning dialog with
   "Continue" and "Cancel" buttons, wait for the user to interact with
   it, and return a value indicating the user's response.  The default
   button is "Continue"; pressing ESC or closing the window is
   equivalent to pressing the "Cancel" button.  See the comment for
   message for more details.  */

extern answer_t warning_continue_cancel (Widget parent,
					 const char *format, ...);


/* Display a formatted string in a Motif warning dialog with "Yes",
   "No", and "Cancel" buttons, wait for the user to interact with it,
   and return a value indicating the user's response.  The default
   button is "Yes"; pressing ESC or closing the window is equivalent
   to pressing the "No" button.  See the comment for message for more
   details.  */

extern answer_t warning_yes_no_cancel (Widget parent,
				       const char *format, ...);


/* Display a formatted string in a Motif information dialog and wait
   for the user to acknowledge it.  See the comment for message for
   more details.  */

extern answer_t information (Widget parent,
			     const char *format, ...);


/* Display a formatted string in a Motif error dialog and wait for the
   user to acknowledge it.  See the comment for message for more
   details.  */

extern answer_t error (Widget parent,
		       const char *format, ...);


/* Display a formatted string in a Motif message dialog box of the
   indicated type, wait for the user to interact with it, and return a
   value indicating the user's response.  The dialog shell is a popup
   child of the parent widget.  The format and variable arguments are
   interpreted as by printf and friends, and the resulting string is
   wrapped to a quarter of the screen width.  The function implements
   its own message loop and will not return until a response from the
   user has been obtained.  */

extern answer_t message (Widget parent, dialog_type_t type,
			 const char *format, ...);


/* Create a string according to the conventions used by printf and
   friends, wrap it to a quarter of the screen width, and use it as
   the message string of the msgbox.  */

extern int msgbox_printf (Widget msgbox, const char *format, ...);


#endif /* DISPLAY_ERROR_H */
